@@ -1,5 +1,5 @@
 /*
- * Copyright 2024 Vsevolod Stakhov
+ * Copyright 2025 Vsevolod Stakhov
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -105,6 +105,7 @@ void rspamd_lua_new_class(lua_State *L,
 	lua_createtable(L, 0, 3 + nmethods);
 
 	if (!seen_index) {
+		/* Default __index = metatable only for plain classes without custom __index */
 		lua_pushstring(L, "__index");
 		lua_pushvalue(L, -2); /* pushes the metatable */
 		lua_settable(L, -3);  /* metatable.__index = metatable */
@@ -985,6 +986,7 @@ rspamd_lua_init(bool wipe_mem)
 	luaopen_tensor(L);
 	luaopen_parsers(L);
 	luaopen_compress(L);
+	luaopen_shingle(L);
 #ifndef WITH_LUAJIT
 	rspamd_lua_add_preload(L, "bit", luaopen_bit);
 	lua_settop(L, 0);
@@ -2400,7 +2402,7 @@ rspamd_lua_try_load_redis(lua_State *L, const ucl_object_t *obj,
 	return FALSE;
 }
 
-void rspamd_lua_push_full_word(lua_State *L, rspamd_stat_token_t *w)
+void rspamd_lua_push_full_word(lua_State *L, rspamd_word_t *w)
 {
 	int fl_cnt;
 
@@ -2520,6 +2522,54 @@ int rspamd_lua_push_words(lua_State *L, GArray *words,
 	return 1;
 }
 
+int rspamd_lua_push_words_kvec(lua_State *L, rspamd_words_t *words,
+							   enum rspamd_lua_words_type how)
+{
+	rspamd_word_t *w;
+	unsigned int i, cnt;
+
+	if (!words || !words->a) {
+		lua_createtable(L, 0, 0);
+		return 1;
+	}
+
+	lua_createtable(L, kv_size(*words), 0);
+
+	for (i = 0, cnt = 1; i < kv_size(*words); i++) {
+		w = &kv_A(*words, i);
+
+		switch (how) {
+		case RSPAMD_LUA_WORDS_STEM:
+			if (w->stemmed.len > 0) {
+				lua_pushlstring(L, w->stemmed.begin, w->stemmed.len);
+				lua_rawseti(L, -2, cnt++);
+			}
+			break;
+		case RSPAMD_LUA_WORDS_NORM:
+			if (w->normalized.len > 0) {
+				lua_pushlstring(L, w->normalized.begin, w->normalized.len);
+				lua_rawseti(L, -2, cnt++);
+			}
+			break;
+		case RSPAMD_LUA_WORDS_RAW:
+			if (w->original.len > 0) {
+				lua_pushlstring(L, w->original.begin, w->original.len);
+				lua_rawseti(L, -2, cnt++);
+			}
+			break;
+		case RSPAMD_LUA_WORDS_FULL:
+			rspamd_lua_push_full_word(L, w);
+			/* Push to the resulting vector */
+			lua_rawseti(L, -2, cnt++);
+			break;
+		default:
+			break;
+		}
+	}
+
+	return 1;
+}
+
 char *
 rspamd_lua_get_module_name(lua_State *L)
 {
@@ -2657,4 +2707,4 @@ int rspamd_lua_geti(lua_State *L, int pos, int i)
 
 	return lua_type(L, -1);
 }
-#endif
\ No newline at end of file
+#endif
