@@ -538,9 +538,8 @@ enum lua_logger_escape_type {
 * @param len
 * @return
 */
-gsize lua_logger_out_type(lua_State *L, int pos, char *outbuf,
-						  gsize len, struct lua_logger_trace *trace,
-						  enum lua_logger_escape_type esc_type);
+gsize lua_logger_out(lua_State *L, int pos, char *outbuf, gsize len,
+					 enum lua_logger_escape_type esc_type);
 
 /**
 * Safely checks userdata to match specified class
@@ -633,7 +632,7 @@ struct rspamd_stat_token_s;
 * @param L
 * @param word
 */
-void rspamd_lua_push_full_word(lua_State *L, struct rspamd_stat_token_s *word);
+void rspamd_lua_push_full_word(lua_State *L, rspamd_word_t *word);
 
 enum rspamd_lua_words_type {
 	RSPAMD_LUA_WORDS_STEM = 0,
@@ -652,6 +651,9 @@ enum rspamd_lua_words_type {
 int rspamd_lua_push_words(lua_State *L, GArray *words,
 						  enum rspamd_lua_words_type how);
 
+int rspamd_lua_push_words_kvec(lua_State *L, rspamd_words_t *words,
+							   enum rspamd_lua_words_type how);
+
 /**
 * Returns newly allocated name for caller module name
 * @param L
