@@ -1,5 +1,5 @@
 /*
- * Copyright 2024 Vsevolod Stakhov
+ * Copyright 2025 Vsevolod Stakhov
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -29,22 +29,123 @@
  * This module hides all complexity: DNS resolving, sessions management, zero-copy
  * text transfers and so on under the hood.
  * @example
+-- Basic GET request with callback
 local rspamd_http = require "rspamd_http"
 
 local function symbol_callback(task)
 	local function http_callback(err_message, code, body, headers)
 		task:insert_result('SYMBOL', 1) -- task is available via closure
+
+		if err_message then
+			-- Handle error
+			return
+		end
+
+		-- Process response
+		if code == 200 then
+			-- Process body and headers
+			for name, value in pairs(headers) do
+				-- Headers are lowercase
+			end
+		end
 	end
 
- 	rspamd_http.request({
- 		task=task,
- 		url='http://example.com/data',
- 		body=task:get_content(),
- 		callback=http_callback,
- 		headers={Header='Value', OtherHeader='Value'},
- 		mime_type='text/plain',
- 		})
- end
+	rspamd_http.request({
+		task=task,
+		url='http://example.com/data',
+		body=task:get_content(),
+		callback=http_callback,
+		headers={Header='Value', OtherHeader='Value', DuplicatedHeader={'Multiple', 'Values'}},
+		mime_type='text/plain',
+	})
+end
+
+-- POST request with JSON body
+local function post_json_example(task)
+	local ucl = require "ucl"
+	local data = {
+		id = task:get_queue_id(),
+		sender = task:get_from()[1].addr
+	}
+
+	local json_data = ucl.to_json(data)
+
+	rspamd_http.request({
+		task = task,
+		url = "http://example.com/api/submit",
+		method = "POST",
+		body = json_data,
+		headers = {['Content-Type'] = 'application/json'},
+		callback = function(err, code, body, headers)
+			if not err and code == 200 then
+				-- Success
+			end
+		end
+	})
+end
+
+-- Synchronous HTTP request (using coroutines)
+local function sync_http_example(task)
+	-- No callback makes this a synchronous call
+	local err, response = rspamd_http.request({
+		task = task,
+		url = "http://example.com/api/data",
+		method = "GET",
+		timeout = 10.0
+	})
+
+	if not err then
+		-- Response is a table with code, content, and headers
+		if response.code == 200 then
+			-- Process response.content
+			return true
+		end
+	end
+	return false
+end
+
+-- Using authentication
+local function auth_example(task)
+	rspamd_http.request({
+		task = task,
+		url = "https://example.com/api/protected",
+		method = "GET",
+		user = "username",
+		password = "secret",
+		callback = function(err, code, body, headers)
+			-- Process authenticated response
+		end
+	})
+end
+
+-- Using HTTPS with SSL options
+local function https_example(task)
+	rspamd_http.request({
+		task = task,
+		url = "https://example.com/api/secure",
+		method = "GET",
+		no_ssl_verify = false, -- Verify SSL (default)
+		callback = function(err, code, body, headers)
+			-- Process secure response
+		end
+	})
+end
+
+-- Using keep-alive and gzip
+local function advanced_example(task)
+	rspamd_http.request({
+		task = task,
+		url = "http://example.com/api/data",
+		method = "POST",
+		body = task:get_content(),
+		gzip = true, -- Compress request body
+		keepalive = true, -- Use keep-alive connection
+		max_size = 1024 * 1024, -- Limit response to 1MB
+		callback = function(err, code, body, headers)
+			-- Process response
+		end
+	})
+end
  */
 
 #define MAX_HEADERS_SIZE 8192
@@ -72,7 +173,16 @@ struct lua_http_cbdata {
 	struct ev_loop *event_loop;
 	struct rspamd_config *cfg;
 	struct rspamd_task *task;
+	lua_State *L;
 	ev_tstamp timeout;
+	/* optional per-request tuning */
+	double connect_timeout;
+	double ssl_timeout;
+	double write_timeout;
+	double read_timeout;
+	double connection_ttl;
+	double idle_timeout;
+	unsigned int max_reuse;
 	struct rspamd_cryptobox_keypair *local_kp;
 	struct rspamd_cryptobox_pubkey *peer_pk;
 	rspamd_inet_addr_t *addr;
@@ -473,6 +583,43 @@ lua_http_make_connection(struct lua_http_cbdata *cbd)
 			cbd->flags |= RSPAMD_LUA_HTTP_FLAG_RESOLVED;
 		}
 
+		/* Optional per-request tuning from table (if present) */
+		if (lua_type(cbd->L, 1) == LUA_TTABLE) {
+			double connect_timeout = 0, ssl_timeout = 0, write_timeout = 0, read_timeout = 0;
+			double connection_ttl = 0, idle_timeout = 0;
+			unsigned int max_reuse = 0;
+			lua_pushstring(cbd->L, "connect_timeout");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) connect_timeout = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			lua_pushstring(cbd->L, "ssl_timeout");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) ssl_timeout = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			lua_pushstring(cbd->L, "write_timeout");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) write_timeout = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			lua_pushstring(cbd->L, "read_timeout");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) read_timeout = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			rspamd_http_connection_set_timeouts(cbd->conn, connect_timeout, ssl_timeout, write_timeout, read_timeout);
+			lua_pushstring(cbd->L, "connection_ttl");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) connection_ttl = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			lua_pushstring(cbd->L, "idle_timeout");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) idle_timeout = lua_tonumber(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			lua_pushstring(cbd->L, "max_reuse");
+			lua_gettable(cbd->L, 1);
+			if (lua_type(cbd->L, -1) == LUA_TNUMBER) max_reuse = lua_tointeger(cbd->L, -1);
+			lua_pop(cbd->L, 1);
+			rspamd_http_connection_set_keepalive_tuning(cbd->conn, connection_ttl, idle_timeout, max_reuse);
+		}
+
 		if (cbd->task) {
 			cbd->conn->log_tag = cbd->task->task_pool->tag.uid;
 
@@ -602,7 +749,7 @@ lua_http_push_headers(lua_State *L, struct rspamd_http_message *msg)
  * @param {string} url specifies URL for a request in the standard URI form (e.g. 'http://example.com/path')
  * @param {function} callback specifies callback function in format  `function (err_message, code, body, headers)` that is called on HTTP request completion. if this parameter is missing, the function performs "pseudo-synchronous" call (see [Synchronous and Asynchronous API overview](/doc/developers/sync_async.html#API-example-http-module)
  * @param {task} task if called from symbol handler it is generally a good idea to use the common task objects: event base, DNS resolver and events session
- * @param {table} headers optional headers in form `[name='value', name='value']`
+ * @param {table} headers optional headers in form `[name='value']` or `[name=['value1', 'value2']]` to duplicate a header with multiple values
  * @param {string} mime_type MIME type of the HTTP content (for example, `text/html`)
  * @param {string/text} body full body content, can be opaque `rspamd{text}` to avoid data copying
  * @param {number} timeout floating point request timeout value in seconds (default is 5.0 seconds)
@@ -612,6 +759,13 @@ lua_http_push_headers(lua_State *L, struct rspamd_http_message *msg)
  * @param {boolean} keepalive enable keep-alive pool
  * @param {string} user for HTTP authentication
  * @param {string} password for HTTP authentication, only if "user" present
+ * @param {number} connect_timeout optional TCP connect timeout (seconds)
+ * @param {number} ssl_timeout optional SSL handshake timeout (seconds)
+ * @param {number} write_timeout optional request write timeout (seconds)
+ * @param {number} read_timeout optional response read timeout (seconds)
+ * @param {number} connection_ttl optional absolute keep-alive connection TTL (seconds)
+ * @param {number} idle_timeout optional keep-alive idle timeout override (seconds)
+ * @param {number} max_reuse optional keep-alive max reuse count per connection
  * @return {boolean} `true`, in **async** mode, if a request has been successfully scheduled. If this value is `false` then some error occurred, the callback thus will not be called.
  * @return In **sync** mode `string|nil, nil|table` In sync mode  error message if any and response as table: `int` _code_, `string` _content_ and `table` _headers_ (header -> value)
  */
@@ -1089,6 +1243,7 @@ lua_http_request(lua_State *L)
 	cbd->url = url;
 	cbd->auth = auth;
 	cbd->task = task;
+	cbd->L = L;
 
 	if (up) {
 		cbd->up = rspamd_upstream_ref(up);
